@@ -1,7 +1,7 @@
 //! `bkt` (pronounced "bucket") is a library for caching subprocess executions. It enables reuse of
 //! expensive invocations across separate processes and supports synchronous and asynchronous
 //! refreshing, TTLs, and other functionality. `bkt` is also a standalone binary for use by shell
-//! scripts and other languages, see https://github.com/dimo414/bkt for binary details.
+//! scripts and other languages, see <https://github.com/dimo414/bkt> for binary details.
 //!
 //! ```no_run
 //! # fn do_something(_: &str) {}
@@ -9,7 +9,7 @@
 //! # use std::time::Duration;
 //! let bkt = bkt::Bkt::in_tmp()?;
 //! let expensive_cmd = bkt::CommandDesc::new(&["wget", "http://example.com"]);
-//! let (result, age) = bkt.execute(&expensive_cmd, Duration::from_secs(3600))?;
+//! let (result, age) = bkt.retrieve(&expensive_cmd, Duration::from_secs(3600))?;
 //! do_something(result.stdout_utf8());
 //! # Ok(()) }
 //! ```
@@ -28,6 +28,15 @@ use anyhow::{Context, Error, Result};
 use serde::{Serialize, Deserialize};
 use serde::de::DeserializeOwned;
 
+#[cfg(feature="debug")]
+macro_rules! debug_msg {
+    ($($arg:tt)*) => { eprintln!("bkt: {}", format!($($arg)*)) }
+}
+#[cfg(not(feature="debug"))]
+macro_rules! debug_msg {
+    ($($arg:tt)*) => {  }
+}
+
 /// Describes a command to be executed and cached. This struct also serves as the cache key.
 /// It consists of a command line invocation and, optionally, a working directory to execute in and
 /// environment variables to set. When set these fields contribute to the cache key, therefore two
@@ -155,7 +164,9 @@ impl CacheKey for CommandDesc {
     fn debug_label(&self) -> Option<String> {
         Some(self.args.iter()
             .map(|a| a.to_string_lossy()).collect::<Vec<_>>().join("-")
-            .chars().filter(|&c| c.is_alphanumeric() || c == '-' || c == '_')
+            .chars()
+            .map(|c| if c.is_whitespace() { '_' } else { c })
+            .filter(|&c| c.is_alphanumeric() || c == '-' || c == '_')
             .take(100).collect())
     }
 }
@@ -180,7 +191,7 @@ mod cmd_tests {
 
     #[test]
     fn debug_label() {
-        assert_eq!(CommandDesc::new(vec!("foo", "bar", "b&r _- a")).debug_label(), Some("foo-bar-br_-a".into()));
+        assert_eq!(CommandDesc::new(vec!("foo", "bar", "b&r _- a")).debug_label(), Some("foo-bar-br__-_a".into()));
     }
 
     #[test]
@@ -343,11 +354,11 @@ trait CacheKey: std::fmt::Debug+Hash+PartialEq {
         if cfg!(feature = "debug") {
             if let Some(label) = self.debug_label() {
                 if !label.is_empty() {
-                    return format!("{}_{:16X}", label, hash);
+                    return format!("{}_{:016X}", label, hash);
                 }
             }
         }
-        format!("{:16X}", hash)
+        format!("{:016X}", hash)
     }
 }
 
@@ -360,7 +371,10 @@ struct CacheEntry<K, V> {
 
 // See https://doc.rust-lang.org/std/fs/fn.soft_link.html
 #[cfg(windows)]
-use std::os::windows::fs::symlink_file as symlink;
+fn symlink<P: AsRef<Path>, Q: AsRef<Path>>(original: P, link: Q) -> Result<()> {
+    std::os::windows::fs::symlink_file(original, link)
+        .context("Windows prevents most programs from creating symlinks; see https://github.com/dimo414/bkt/issues/3")
+}
 #[cfg(unix)]
 use std::os::unix::fs::symlink;
 
@@ -432,6 +446,7 @@ impl Cache {
         let file = File::open(&path);
         if let Err(ref e) = file {
             if e.kind() == ErrorKind::NotFound {
+                debug_msg!("lookup {} not found", path.display());
                 return Ok(None);
             }
         }
@@ -444,13 +459,16 @@ impl Cache {
         let mtime = std::fs::metadata(&path)?.modified()?;
         let elapsed = mtime.elapsed();
         if elapsed.is_err() || elapsed.unwrap() > max_age {
+            debug_msg!("lookup {} expired", path.display());
             std::fs::remove_file(&path).context("Failed to remove expired data")?;
             return Ok(None);
         }
         // Ignore false-positive hits that happened to collide with the hash code
         if &found.key != key {
+            debug_msg!("lookup {} hash collision", path.display());
             return Ok(None);
         }
+        debug_msg!("lookup {} found", path.display());
         Ok(Some((found.value, mtime)))
     }
 
@@ -478,12 +496,15 @@ impl Cache {
         let file = OpenOptions::new().create_new(true).write(true).open(&path)?;
         let entry = CacheEntry{ key, value };
         Cache::serialize(BufWriter::new(&file), &entry).context("Serialization failed")?;
+        debug_msg!("store data {}", path.display());
         // Roundabout approach to an atomic symlink replacement
         // https://github.com/dimo414/bash-cache/issues/26
         let tmp_symlink = Cache::rand_filename(&self.key_dir(), "tmp-symlink");
         // Note: this will fail if filename collides, could retry in a loop if that happens
         symlink(&path, &tmp_symlink)?;
-        std::fs::rename(&tmp_symlink, self.key_path(&entry.key.cache_key()))?;
+        let key_path = self.key_path(&entry.key.cache_key());
+        std::fs::rename(&tmp_symlink, &key_path)?;
+        debug_msg!("store key {}", key_path.display());
         Ok(())
     }
 
@@ -502,12 +523,14 @@ impl Cache {
             let last_attempt_file = self.cache_dir.join("last_cleanup");
             if let Ok(metadata) = last_attempt_file.metadata() {
                 if metadata.modified()?.elapsed()? < Duration::from_secs(30) {
+                    debug_msg!("cleanup skip recent");
                     return Ok(());
                 }
             }
             File::create(&last_attempt_file)?; // resets mtime if already exists
 
             // First delete stale data files
+            debug_msg!("cleanup data {}", &self.data_dir().display());
             if let Ok(data_dir_iter) = std::fs::read_dir(&self.data_dir()) {
                 for entry in data_dir_iter {
                     let ttl_dir = entry?.path();
@@ -525,6 +548,7 @@ impl Cache {
             }
 
             // Then delete broken symlinks
+            debug_msg!("cleanup keys {}", &self.key_dir().display());
             if let Ok(key_dir_iter) = std::fs::read_dir(&self.key_dir()) {
                 for entry in key_dir_iter {
                     let symlink = entry?.path();
@@ -679,19 +703,26 @@ mod cache_tests {
 
 /// This struct is the main API entry point for the `bkt` library, allowing callers to invoke and
 /// cache subprocesses for later reuse.
-#[derive(Debug)]
+#[derive(Clone, Debug)]
 pub struct Bkt {
     cache: Cache,
+    cleanup_on_refresh: bool,
+    persist_failures: bool,
 }
 
 impl Bkt {
-    /// Creates a new Bkt instance using the [`std::env::temp_dir`] as the cache location.
+    fn temp_dir() -> PathBuf {
+        std::env::var_os("BKT_TMPDIR").map(PathBuf::from).unwrap_or_else(std::env::temp_dir)
+    }
+
+    /// Creates a new Bkt instance using the [`std::env::temp_dir`] as the cache location. If a
+    /// `BKT_TMPDIR` environment variable is set that value will be preferred.
     ///
     /// # Errors
     ///
     /// If preparing the tmp cache directory fails.
     pub fn in_tmp() -> Result<Self> {
-        Bkt::create(std::env::temp_dir())
+        Bkt::create(Bkt::temp_dir())
     }
 
     /// Creates a new Bkt instance.
@@ -711,6 +742,8 @@ impl Bkt {
         Bkt::restrict_dir(&cache_dir)?;
         Ok(Bkt {
             cache: Cache::new(&cache_dir),
+            cleanup_on_refresh: true,
+            persist_failures: true,
         })
     }
 
@@ -724,8 +757,32 @@ impl Bkt {
         self
     }
 
+    /// By default a background cleanup thread runs on cache misses and calls to [`Bkt::refresh()`]
+    /// to remove stale data. You may prefer to manage cleanup yourself if you expect frequent cache
+    /// misses and want to minimize the number of threads being created. See [`Bkt::cleanup_once()`]
+    /// and [`Bkt::cleanup_thread()`] if you set this to `false`.
+    pub fn cleanup_on_refresh(mut self, cleanup: bool) -> Self {
+        self.cleanup_on_refresh = cleanup;
+        self
+    }
+
+    /// Configures this instance to not cache invocations that return non-zero exit codes. This only
+    /// affects _writing_ to the cache; if a failed invocation has already been cached (e.g. by a
+    /// different instance) that data will still be used until it expires.
+    ///
+    /// **WARNING:** use this function with caution. Discarding invocations that fail can overload
+    /// downstream resources that were protected by the caching layer limiting QPS. For example,
+    /// if a website is rejecting a fraction of requests to shed load and then clients start
+    /// sending _more_ requests when their attempts fail the website could be taken down outright by
+    /// the added load. In other words, using this function can lead to accidental DDoSes.
+    pub fn discard_failures(mut self, discard_failures: bool) -> Self {
+        // Flip the boolean here to make the conditional in retrieve() clearer
+        self.persist_failures = !discard_failures;
+        self
+    }
+
     #[cfg(not(unix))]
-    fn restrict_dir(cache_dir: &Path) -> Result<()> { Ok(()) }
+    fn restrict_dir(_cache_dir: &Path) -> Result<()> { Ok(()) }
     #[cfg(unix)]
     fn restrict_dir(cache_dir: &Path) -> Result<()> {
         use std::os::unix::fs::PermissionsExt;
@@ -760,51 +817,26 @@ impl Bkt {
     /// than the given TTL.
     ///
     /// If stale or not found the command is executed and the result is cached and then returned.
-    /// A zero-duration age will be returned if this invocation refreshed the cache. A background
-    /// cleanup thread will also run on cache misses to remove stale data.
+    /// A zero-duration age will be returned if this invocation refreshed the cache.
     ///
     /// # Errors
     ///
     /// If looking up, deserializing, executing, or serializing the command fails. This generally
     /// reflects a user error such as an invalid command.
-    // TODO better name than execute?
     // TODO per C-CALLER-CONTROL perhaps this should consume the CommandDesc rather than cloning it
-    // in execute_subprocess(). See https://rust-lang.github.io/api-guidelines/flexibility.html
-    // See also C-BUILDER in https://rust-lang.github.io/api-guidelines/type-safety.html
-    pub fn execute(&self, command: &CommandDesc, ttl: Duration) -> Result<(Invocation, Duration)> {
-        self._execute(command, ttl, true)
-    }
-
-    /// See the documentation on `execute()`. This functions like `execute()` but does not attempt
-    /// to clean up stale data. Prefer this method if you decide to manage cleanup yourself via
-    /// `cleanup_once()` or `cleanup_thread()`.
-    ///
-    /// # Errors
-    ///
-    /// If looking up, deserializing, executing, or serializing the command fails. This generally
-    /// reflects a user error such as an invalid command.
-    pub fn execute_without_cleanup(&self, command: &CommandDesc, ttl: Duration) -> Result<(Invocation, Duration)> {
-        self._execute(command, ttl, false)
-    }
-
-    fn _execute(&self, command: &CommandDesc, ttl: Duration, cleanup: bool) -> Result<(Invocation, Duration)> {
+    //     in execute_subprocess(). See https://rust-lang.github.io/api-guidelines/flexibility.html
+    //     See also C-BUILDER in https://rust-lang.github.io/api-guidelines/type-safety.html
+    pub fn retrieve(&self, command: &CommandDesc, ttl: Duration) -> Result<(Invocation, Duration)> {
         let cached = self.cache.lookup(command, ttl).context("Cache lookup failed")?;
         let result = match cached {
             Some((cached, mtime)) => (cached, mtime.elapsed()?),
             None => {
-                let mut cleanup_hook = None;
-                if cleanup {
-                    // clean the cache in the background on a cache-miss; this will usually
-                    // be much faster than the actual background process.
-                    cleanup_hook = Some(self.cleanup_once());
-                }
+                let cleanup_hook = self.maybe_cleanup_once();
                 let result = Bkt::execute_subprocess(command).context("Subprocess execution failed")?;
-                self.cache.store(command, &result, ttl).context("Cache write failed")?;
-                if let Some(cleanup_hook) = cleanup_hook {
-                    if let Err(e) = cleanup_hook.join().expect("cleanup thread panicked") {
-                        eprintln!("bkt: cache cleanup failed: {:?}", e);
-                    }
+                if self.persist_failures || result.exit_code == 0 {
+                    self.cache.store(command, &result, ttl).context("Cache write failed")?;
                 }
+                Bkt::join_cleanup_thread(cleanup_hook);
                 (result, Duration::default())
             }
         };
@@ -819,11 +851,33 @@ impl Bkt {
     /// If executing or serializing the command fails. This generally reflects a user error such as
     /// an invalid command.
     pub fn refresh(&self, command: &CommandDesc, ttl: Duration) -> Result<Invocation> {
+        let cleanup_hook = self.maybe_cleanup_once();
         let result = Bkt::execute_subprocess(command).context("Subprocess execution failed")?;
-        self.cache.store(command, &result, ttl).context("Cache write failed")?;
+        if self.persist_failures || result.exit_code == 0 {
+            self.cache.store(command, &result, ttl).context("Cache write failed")?;
+        }
+        Bkt::join_cleanup_thread(cleanup_hook);
         Ok(result)
     }
 
+    /// Clean the cache in the background on a cache-miss; this will usually
+    /// be much faster than the actual background process.
+    fn maybe_cleanup_once(&self) -> Option<std::thread::JoinHandle<Result<()>>> {
+        if self.cleanup_on_refresh {
+            Some(self.cleanup_once())
+        } else {
+            None
+        }
+    }
+
+    fn join_cleanup_thread(cleanup_hook: Option<std::thread::JoinHandle<Result<()>>>) {
+        if let Some(cleanup_hook) = cleanup_hook {
+            if let Err(e) = cleanup_hook.join().expect("cleanup thread panicked") {
+                eprintln!("bkt: cache cleanup failed: {:?}", e);
+            }
+        }
+    }
+
     /// Initiates a single cleanup cycle of the cache, removing stale data in the background. This
     /// should be invoked by short-lived applications early in their lifecycle and then joined
     /// before exiting. `execute_and_cleanup` can be used instead to only trigger a cleanup on a
@@ -864,6 +918,16 @@ mod bkt_tests {
     use super::*;
     use test_dir::{TestDir, DirBuilder, FileType};
 
+    // Just validating that Bkt can be cloned to create siblings with different settings.
+    #[test]
+    #[allow(clippy::redundant_clone)]
+    fn cloneable() {
+        let dir = TestDir::temp();
+        let bkt = Bkt::create(dir.path("cache")).unwrap();
+        let _scoped = bkt.clone().scoped("scope".into());
+        let _no_cleanup = bkt.clone().cleanup_on_refresh(false);
+    }
+
     #[test]
     fn cached() {
         let dir = TestDir::temp();
@@ -871,21 +935,53 @@ mod bkt_tests {
         let cmd = CommandDesc::new(
             vec!("bash", "-c", "echo \"$RANDOM\" > \"${1:?}\"; cat \"${1:?}\"", "arg0", file.to_str().unwrap()));
         let bkt = Bkt::create(dir.path("cache")).unwrap();
-        let (first_inv, _) = bkt.execute(&cmd, Duration::from_secs(10)).unwrap();
+        let (first_inv, _) = bkt.retrieve(&cmd, Duration::from_secs(10)).unwrap();
 
         for _ in 1..3 {
-            let (subsequent_inv, _) = bkt.execute(&cmd, Duration::from_secs(10)).unwrap();
+            let (subsequent_inv, _) = bkt.retrieve(&cmd, Duration::from_secs(10)).unwrap();
             assert_eq!(first_inv, subsequent_inv);
         }
     }
 
+    #[test]
+    fn discard_failures() {
+        let dir = TestDir::temp();
+        let output = dir.path("output");
+        let code = dir.path("code");
+
+        let cmd = CommandDesc::new(
+            vec!("bash", "-c", "cat \"${1:?}\"; exit \"$(< \"${2:?}\")\"", "arg0", output.to_str().unwrap(), code.to_str().unwrap()));
+        let bkt = Bkt::create(dir.path("cache")).unwrap().discard_failures(true);
+
+        write!(File::create(&output).unwrap(), "A").unwrap();
+        write!(File::create(&code).unwrap(), "10").unwrap();
+        let (first_inv, _) = bkt.retrieve(&cmd, Duration::from_secs(10)).unwrap();
+        assert_eq!(first_inv.exit_code, 10, "{:?}\nstderr:{}", first_inv, first_inv.stderr_utf8());
+        assert_eq!(first_inv.stdout_utf8(), "A");
+
+        write!(File::create(&output).unwrap(), "B").unwrap();
+        let (subsequent_inv, _) = bkt.retrieve(&cmd, Duration::from_secs(10)).unwrap();
+        // call is not cached
+        assert_eq!(subsequent_inv.stdout_utf8(), "B");
+
+        write!(File::create(&output).unwrap(), "C").unwrap();
+        write!(File::create(&code).unwrap(), "0").unwrap();
+        let (success_inv, _) = bkt.retrieve(&cmd, Duration::from_secs(10)).unwrap();
+        assert_eq!(success_inv.exit_code, 0);
+        assert_eq!(success_inv.stdout_utf8(), "C");
+
+        write!(File::create(&output).unwrap(), "D").unwrap();
+        let (cached_inv, _) = bkt.retrieve(&cmd, Duration::from_secs(10)).unwrap();
+        assert_eq!(success_inv, cached_inv);
+    }
+
     #[test]
     fn with_working_dir() {
         let dir = TestDir::temp().create("dir", FileType::Dir);
         let cwd = dir.path("dir");
         let cmd = CommandDesc::new(vec!("bash", "-c", "echo Hello World > file")).with_working_dir(&cwd);
         let bkt = Bkt::create(dir.path("cache")).unwrap();
-        let (result, _) = bkt.execute(&cmd, Duration::from_secs(10)).unwrap();
+        let (result, _) = bkt.retrieve(&cmd, Duration::from_secs(10)).unwrap();
         assert_eq!(result.stderr_utf8(), "");
         assert_eq!(result.exit_code(), 0);
         assert_eq!(std::fs::read_to_string(cwd.join("file")).unwrap(), "Hello World\n");
@@ -900,9 +996,9 @@ mod bkt_tests {
         let dir = TestDir::temp().create("dir", FileType::Dir);
         let cmd = CommandDesc::new(vec!("bash", "-c", "echo \"FOO:${FOO:?}\"")).with_env_value("FOO", "bar");
         let bkt = Bkt::create(dir.path("cache")).unwrap();
-        let (result, _) = bkt.execute(&cmd, Duration::from_secs(10)).unwrap();
+        let (result, _) = bkt.retrieve(&cmd, Duration::from_secs(10)).unwrap();
         assert_eq!(result.stderr_utf8(), "");
         assert_eq!(result.exit_code(), 0);
         assert_eq!(result.stdout_utf8(), "FOO:bar\n");
     }
-}
\ No newline at end of file
+}
